import { BehaviorSubject, Observable, Subject, Unsubscribable } from "rxjs"
import { WrappedMessage } from "../utils/message.ordering"

/* EVENT BUS */
export interface GeneralEvent<T> {
    id: string,
    type: EventType,
    event: EventMessage,
    date: Date,
    data: T
    transport?: Transport,
}

/* MANAGEMENT */
export interface MessageTransmissionManagerInterface {
    subscribeForTransmission(): Observable<GeneralEvent<any>>
}

export interface AdapterManagerInterface {
    subscribeForAdapters(): Observable<GeneralEvent<any>>
}


/* TRANSMISSION COMPONENTS */
export interface MessageTransmissionInterface {
}

export interface MessageReceiverInterface extends MessageTransmissionInterface {
    getIncoming(): Observable<any>
}

export interface MessageTransmitterInterface extends MessageTransmissionInterface {
    emit(message: any): void
}

export interface MessageRequestResponseInterface extends MessageTransmissionInterface {
    send(message: any): Observable<any>
}

/* ADAPTER COMPONENTS */
export interface AdapterInterface {
    adapterId: string,
    role: TransmissionRole,
    transport: Transport
}

export interface TransmitterAdapterInterface extends AdapterInterface {
    emit(message: WrappedMessage): void
}

export interface ReceiverAdapterInterface extends AdapterInterface {
    subscribeForIncoming(): Observable<any>
}


/* Utils */
export interface TransmissionProfile {
    id: string,
    name: string,
    dateCreated: Date
}

export interface TransmitterProfile extends TransmissionProfile {
}

export interface ReceiverProfile extends TransmissionProfile {
}

export interface RequestResponseProfile extends TransmissionProfile {
}

export interface TransportMessage {
    id: string,
    dateCreated: Date,
    transport: Transport,
    target?: string,
    payload: any
}

export interface FisMessage {
    header: {
        messageID: string,
        messageName: `NotificationMessage` | `ResponseMessage` | `RequestMessage`
    },
    data: any
}

export interface TransmissionInterface {
    clientId: string,
    transmitter: MessageTransmitterInterface,
    receiver: MessageReceiverInterface,
    requestResponse: MessageRequestResponseInterface,
    event: Observable<GeneralEvent<TransmissionInterface>>
}

export type Transport = 'Websocket' | 'Http' | 'TCP' | undefined

export enum AdapterTransmissionRole {
    Transmitter,
    Receiver,
    RequestResponse
}


export type EventType = `General Event` | 'Transport Event' | 'Transmission Event' | 'Adapter Event'
export type TransmissionRole = `Transmitter` | 'Receiver' | 'RequestResponse'
export type EventMessage = 'Server Started' | 'New Client' | 'Client Connected' | 'Client Re-connected' | 'Client Disconnected' | `Server Disconnected` | 'New Message' | `Notification` | `New Server` | `Server Connected` | `New Transport` | 'New Adapter' | 'Re-Flush' | 'New Transport' | 'New Transmission' | 'Adapter Manager Started' | 'Available Transport'
export type ConnectionState = 'ONLINE' | 'OFFLINE'

export interface TransportServiceInterface {
    getInfo(): TransportServiceProfile
    emit(message: TransportMessage): void
    subscribeForTransportEvent(): Observable<GeneralEvent<any>> //all messages and whatever event will go through this, easier to implemnet across different transport protocol
}

export interface TransportServiceProfile {
    transportServiceId: string,
    transport: Transport
}

export interface Info {
    transport: Transport
}

export interface ClientObject {
    clientId: string,
    dateCreated: Date,
    connectionState: BehaviorSubject<ConnectionState>,
    transport: Transport,
    transportServiceId: string 
}


export interface AdapterProfile {
    id: string,
    transportType: Transport,
}

export interface TransportSet {
    transport: Transport,
    port: number
}
export interface TransportProfileMessage {
    clientId: string,
    message?: string,
    payload?: any
}