import { Socket as ClientSocket, io } from 'socket.io-client'
import { Server, Socket as SocketForConnectedClient } from "socket.io"
import { Observable, Observer, Subject } from "rxjs";
import { createServer, request as httpRequest } from "http";
import express, { Response } from 'express';
import { Express } from 'express';
import { postAxiosRequest } from '../utils/http.utils';
import axios from 'axios';
let fromServer = new Subject<{ event: 'profile' | 'message', payload: any }>()
let toServer = new Subject<{ event: 'profile' | 'message', payload: any }>()

startSocketServer(3001)
// startSocketServer(3002)
startClientSocketConnection('http://localhost:3000')
// startHttpServer(3001).then((app: Express) => {
//     operateHttpServer(app, 'http://localhost:3000/')
// })
consoleLog()

function consoleLog(): void {
    fromServer.subscribe(message => console.log(`From Server`, message.event, message.payload?.thisMessageID ?? 'undefined message'))
    toServer.subscribe(message => console.log(`To Server`, message.event, message.payload?.thisMessageID ?? 'undefined message'))
}

function startSocketServer(port: number): void {
    console.log(`Socket Server ${port} Started....`)
    let httpServer = createServer();
    let socketServer = new Server(httpServer)

    socketServer.on('connection', (socket: SocketForConnectedClient) => {
        socket.on(`profile`, (msg) => {
            toServer.next({ event: 'profile', payload: msg })
        })

        socket.on('message', (msg) => {
            toServer.next({ event: 'message', payload: msg })
        })

        socket.on(`disconnect`, () => {
            console.log(`I shouldn't be here lol`)
        })

        fromServer.subscribe(message => {
            socketServer.emit(message.event, message.payload)
        })
    })

    socketServer.engine.on("connection_error", (err) => {
        console.log(err.req);      // the request object
        console.log(err.code);     // the error code, for example 1
        console.log(err.message);  // the error message, for example "Session ID unknown"
        console.log(err.context);  // some additional error context
    });

    // Start the HTTP server on 127.0.0.1 with the given port
    httpServer.listen(port, '0.0.0.0', () => {
        console.log(`Socket server listening on ${port}`);
    });

}


function startClientSocketConnection(serverUrl: string): void {
    // let clientSocket = io(serverUrl)
    let clientSocket: ClientSocket = io(serverUrl, {
        reconnection: true,              // Enable automatic reconnections
        reconnectionAttempts: 1000,       // Retry up to 10 times
        reconnectionDelay: 500,          // Start with a 500ms delay
        reconnectionDelayMax: 10000,     // Delay can grow to a max of 10 seconds
        randomizationFactor: 0.3,
    })

    toServer.subscribe(message => { clientSocket.emit(message.event, message.payload) })

    clientSocket.on(`connect`, () => {

    })

    clientSocket.on(`profile`, (msg) => {
        fromServer.next({ event: 'profile', payload: msg })
    })

    clientSocket.on(`message`, (msg) => {
        fromServer.next({ event: 'message', payload: msg })
    })

    clientSocket.on(`disconnect`, () => {

    })
}

async function startHttpServer(port: number): Promise<Express> {
    return new Promise((resolve, reject) => {
        let app: Express = express();
        // Middleware to parse JSON requests
        app.use(express.json());
        app.listen(port, () => {
            console.log({ message: `Server running at http://localhost:${port}` });
        });
        resolve(app)
    })
}

function operateHttpServer(app: Express, url: string): void {
    app.post('/profile', (req, res) => {
        postAxiosRequest(url + `profile`, req.body).then((response) => {
            res.json(response)
        }).catch((error) => {
            console.log(error)
            res.json(error)
        })
    })

    app.post('/message', (req, res) => {
        postAxiosRequest(url + `message`, req.body).then((response) => {
            console.log(response)
            res.json(response)
        }).catch((error) => {
            console.log(error)
            res.json(error)
        })
    })

    app.get('/poll', (req, res) => {
        console.log('Client connected for long polling.');
        // Flag to track if the response has been sent
        let responseSent = false;
        // Subscribe to the data stream
        const subscription = handleClientHttpConnection(url).subscribe({
            next: (message: any) => {
                if (!responseSent) {
                    console.log(`Sending data to client: ${JSON.stringify(message)}`);
                    res.json(message); // Send the data to the client
                    responseSent = true; // Mark response as sent
                    subscription.unsubscribe(); // Unsubscribe to close this request
                }
            },
            error: (err: any) => {
                if (!responseSent) {
                    console.error('Error in data stream:');
                    res.status(500).send('Internal Server Error');
                    responseSent = true; // Mark response as sent
                }
                subscription.unsubscribe(); // Ensure cleanup
            },
            complete: () => {
                if (!responseSent) {
                    console.log('Data stream completed.');
                    res.status(204).send(); // No Content
                    responseSent = true; // Mark response as sent
                }
                subscription.unsubscribe(); // Ensure cleanup
            },
        });

        // Timeout if no data is emitted within a specified duration
        const timeout = setTimeout(() => {
            if (!responseSent) {
                console.log({ message: 'No data emitted. Sending timeout response.' });
                res.status(204).send(); // No Content
                responseSent = true; // Mark response as sent
                subscription.unsubscribe(); // Ensure cleanup
            }
        }, 15000); // 15 seconds timeout (adjust as needed)

        // Handle client disconnection
        res.on('close', () => {
            if (!responseSent) {
                console.error(`Http Client disconnected`);

                subscription.unsubscribe(); // Ensure cleanup
            }
            clearTimeout(timeout); // Clear timeout to avoid unnecessary execution
        });
    });
}


// For client usage
export function handleClientHttpConnection(url: string): Observable<any> {
    return new Observable((eventNotification: Observer<any>) => {
        let active: boolean = true; // Flag to control polling lifecycle

        const longPoll = async () => {
            while (active) {
                try {
                    // Axios request with timeout
                    const response = await axios.get(`${url}poll`); // removing the timeout temporarily. 
                    if (response.status === 200) {
                        const data = response.data;
                        eventNotification.next(data)
                    } else if (response.status === 204) {
                        console.log('No new messages from the server.');
                    } else {
                        throw new Error(`Unexpected response status: ${response.status}`);
                    }
                } catch (error: unknown) {
                    console.error(`Unknown Error.`) // culprit is here
                    // Error handling with server disconnect notification
                    let errorMessage: string;

                    if (axios.isAxiosError(error)) {
                        if (error.response) {
                            errorMessage = `Server returned status ${error.response.status}: ${error.response.statusText}`;
                        } else if (error.code === 'ECONNABORTED') {
                            errorMessage = 'Request timed out.';
                        } else {
                            errorMessage = error.message || 'An Axios error occurred.';
                        }
                    } else if (error instanceof Error) {
                        errorMessage = error.message;
                    } else {
                        errorMessage = 'An unknown error occurred during polling.';
                    }

                    console.error(`Polling error: ${errorMessage}`);
                    // observer.error(new Error(errorMessage)); // Notify subscribers of the error
                    break; // Stop polling on error
                }
            }
        };

        longPoll();

        // Cleanup logic for unsubscribing
        return () => {
            console.log({ message: 'Unsubscribed from the long-polling channel.' });
            eventNotification.complete(); // Notify completion
        };
    });
}