import { Observable, Subject } from "rxjs";
import { v4 as uuidv4 } from 'uuid'
import { Socket as SocketForConnectedServer } from 'socket.io-client'
import { Socket as SocketForConnectedClient } from "socket.io"
import { handleClientSocketConnection, handleNewSocketClient, startClientSocketConnection, startSocketServer } from "../utils/socket.utils";
import { WrappedMessage } from "../utils/message.ordering";
import ConsoleLogger from "../utils/log.utils";
import { ClientObject, FisMessage, GeneralEvent, Transport, TransportMessage, TransportServiceInterface, TransportServiceProfile } from "../interface/interface";

/* Just code in the context that this websocket service will be handling multiple UI clients. Can think about the server communication at a later time. */
export class WebsocketTransportService implements TransportServiceInterface {
    private console: ConsoleLogger = new ConsoleLogger(`WebsocketTransportService`, ['transport'])
    private info!: TransportServiceProfile
    private connectedSocketServer: ConnectedSocketServer[] = [] // to allow the possibility of having to communicate with multiple servers as a client
    private connectedClientSocket: ConnectedSocketClient[] = [] // to keep track of the all the clients that are connected
    // private incomingMessage: Subject<TransportMessage> = new Subject() // this is only for client roles only atm
    private transportEvent!: Subject<GeneralEvent<any>>

    constructor(event: Subject<GeneralEvent<any>>) {
        this.console.log({ message: `WebsocketTransportService: Constructing socket transport service....` })
        this.info = {
            transportServiceId: uuidv4(),
            transport: 'Websocket'
        }
        // logic here
        this.transportEvent = event
    }

    public startServer(port: number): void {
        startSocketServer(port).subscribe({
            next: (connectedClient: SocketForConnectedClient) => {
                handleNewSocketClient(this.info.transportServiceId, connectedClient, this.connectedClientSocket).subscribe({
                    next: event => this.transportEvent.next(event),
                    error: error => this.console.error({ message: `Observer Error: ${error}`, details: error }),
                    complete: () => this.console.log({ message: `Client ${connectedClient.id} disconnected...` })
                })
            },
            error: error => this.console.error({ message: `Observer Error`, details: error }),
            complete: () => this.console.log({ message: `Not sure what this is for...` })
        })
    }

    public startClient(url: string): void {
        // logic here
        startClientSocketConnection(url).then((socket: SocketForConnectedServer) => {
            handleClientSocketConnection(this.info.transportServiceId, socket, this.connectedSocketServer).subscribe(this.transportEvent)
        }).catch((error) => {
            this.console.log({ message: `Observer Error`, details: error })
        })
    }


    public emit(message: TransportMessage): void {
        this.console.log({ message: `Emitting: ${((message.payload as WrappedMessage).payload as FisMessage).header.messageID} to ${message.target}`, details: message })
        let clientObj: ConnectedSocketClient | undefined = this.connectedClientSocket.find(obj => obj.clientId === message.target)
        let serverObj: ConnectedSocketServer | undefined = this.connectedSocketServer.find(obj => obj.clientId === message.target)
        // this.console.log({ message: `${serverObj?.connectionState.getValue(), serverObj?.id}` })
        // for server usage
        if (clientObj && clientObj.connectionState.getValue() == 'ONLINE') {
            clientObj.socketInstance.emit(`message`, message.payload)
        }
        // for client usage
        if (serverObj && serverObj.connectionState.getValue() == 'ONLINE') {
            serverObj.socketInstance.emit(`message`, message.payload)
        }
    }

    public subscribeForTransportEvent(): Observable<GeneralEvent<any>> {
        return this.transportEvent.asObservable()
    }

    public getInfo(): TransportServiceProfile {
        return this.info
    }
}

export interface ConnectedSocketClient extends ClientObject {
    socketInstance: SocketForConnectedClient
}

export interface ConnectedSocketServer extends ClientObject {
    socketInstance: SocketForConnectedServer
}
